from click import BadParameter
import re
import typer
from typing import Tuple, List, Dict, Optional
from pathlib import Path
from dovado_rtl.doc_parsing import get_directives, get_directives_paragraph
from dovado_rtl.src_parsing import SourceParser
from dovado_rtl.vivado_interaction import get_parts
from dovado_rtl.antlr.hdl_representation import PortDirectionEnum
from dovado_rtl.simple_types import Metric
from movado import models, controllers


def validate_power_of_2(value: List[str], ctx: typer.Context):
    if len(ctx.params["parameters"]) != len(value):
        raise BadParameter(
            "power of 2 constraint must be specified for each parameter, mismatching list lengths"
            + ", parameter list lenght:"
            + str(len(ctx.params["parameters"]))
            + ", constraint list length: "
            + str(len(value))
        )
    for i in value:
        if i not in {"y", "n"}:
            raise BadParameter("Accepted input are only y/n")
    return value


def validate_int_metrics(value: List[int]):
    for i in value:
        if i < 0:
            raise BadParameter(
                "Invalid metric '" + str(i) + ", all parameters must be >= 0"
            )
    return value


def parse_comma_separated_list(
    regexp_element: str,
    to_parse: str,
) -> Optional[List[str]]:
    return (
        None
        if not re.fullmatch(
            "^("
            + regexp_element
            + ")"
            + "(([ \t]*,[ \t]*)("
            + regexp_element
            + "))*$",
            to_parse,
        )
        else [i for i in re.sub(r"[ \t]*", "", to_parse).split(",")]
    )


def ask_utilization_metrics(
    util_indices: Dict[str, List[str]],
    pre_selected_metrics: Optional[List[int]] = None,
) -> List[Metric]:
    info_prompt = "Percentage Utilisation indices available:\n"
    i = 0
    counter_dict = {}
    for section in util_indices.keys():
        info_prompt += section + "\n"
        for util_metric in util_indices[section]:
            i = i + 1
            counter_dict[i] = Metric(
                utilisation=(section, util_metric), is_frequency=False
            )
            info_prompt += "\t(" + str(i) + ") " + util_metric + "\n"

    while True:
        if not pre_selected_metrics:
            typer.echo(info_prompt)
            user_input = typer.prompt(
                "Enter a comma-separated list of percentage "
                + "indices to optimize, input 0 to select frequency"
            )
        try:
            if not pre_selected_metrics:
                parsed_list = [
                    int(i)
                    for i in parse_comma_separated_list(r"[+]?\d+", user_input)
                ]
            else:
                parsed_list = pre_selected_metrics
                pre_selected_metrics = None
            if parsed_list and (
                parsed_list == []
                or max([i for i in parsed_list if i > 0]) < len(counter_dict)
            ):
                out: List[Metric] = [
                    counter_dict[i] for i in {i for i in parsed_list if i > 0}
                ]
                return (
                    (out + [Metric(None, True)]) if (0 in parsed_list) else out
                )
        except Exception as e:
            print(e)
            print(
                "Invalid input, please enter"
                + " a comma separated list of integer numbers"
            )


def validate_nearest_distance(value: int) -> int:
    if value < 0:
        raise typer.BadParameter(
            "Invalid distance value: "
            + str(value)
            + " non-negative integer expected"
        )
    return value


def validate_estimator(value: str) -> str:
    if value in models:
        return value
    raise typer.BadParameter(
        "Invalid estimation model "
        + value
        + " available models are: "
        + str(models)
    )


def validate_controller(value: str) -> str:
    if value in controllers:
        return value
    raise typer.BadParameter(
        "Invalid controller "
        + value
        + " available controllers are: "
        + str(controllers)
    )


def validate_board(value: str) -> str:
    parts = get_parts()
    if value in set(parts.keys()):
        return value
    raise typer.BadParameter(
        "Invalid board " + value + " available boards are: " + str(parts)
    )


def validate_parameters(ctx: typer.Context, value: List[str]) -> List[str]:

    file_path = Path(ctx.params["file_path"]).absolute()
    if not ctx.obj:
        parsed_src = SourceParser(
            str(SourceParser.get_project_path(file_path)), str(file_path)
        )
        parsed_src.set_entity(str(parsed_src.get_entities()[0].get_name()))
        ctx.obj = parsed_src
    else:
        parsed_src: SourceParser = ctx.obj
    parameter_name_list = [p.get_name() for p in parsed_src.get_parameters()]
    for chosen_parameter in value:
        if chosen_parameter not in parameter_name_list:
            raise typer.BadParameter(
                "Chosen parameter: "
                + chosen_parameter
                + " is not among available parameters "
                + str(parameter_name_list)
            )
    return value


def validate_clock_port(ctx: typer.Context, value: str) -> str:
    parsed_src: SourceParser = ctx.obj
    for port in parsed_src.get_ports():
        if (
            port.get_name() == value
            and port.get_direction().direction is PortDirectionEnum.INPUT
        ):
            return value
    raise typer.BadParameter(
        "Chosen clock port: "
        + value
        + " is not among available input ports "
        + str(
            [
                p.get_name()
                for p in parsed_src.get_ports()
                if p.get_direction().direction is PortDirectionEnum.INPUT
            ]
        )
    )


def validate_directives(value: Tuple[str, str, str]):
    commands = ["synthesis", "place", "route"]
    for c in commands:
        directive = value[commands.index(c)]
        directives = get_directives((get_directives_paragraph(c)))
        if directive not in directives:
            raise typer.BadParameter(
                "Chosen directive for "
                + c
                + ": "
                + directive
                + " is not among available directives for "
                + c
                + " "
                + str(directives)
            )
    return value


def validate_target_clock(value: float) -> float:
    if value > 0:
        return value
    raise BadParameter(
        "Target clock must be positive, invalid target clock input: "
        + str(value)
    )


def validate_param_ranges(value: List[int]) -> List[int]:
    if len(value) % 2 == 0:
        return value
    raise typer.BadParameter(
        "Parameter ranges list must be of even length, "
        + "each odd element is the start of the range and the next one is the end of the range"
    )



import abc
from abc import abstractmethod
from typing import Tuple, List, Dict, Optional
from dovado_rtl.antlr.hdl_representation import (
    Entity,
    Parameter,
    TopLevel,
    Port,
)
from dovado_rtl.enums import RTL
from dovado_rtl.simple_types import DesignValue, Example, Metric
from dovado_rtl.fill_handler import FillHandler


class AbstractEstimator(abc.ABC):
    @abstractmethod
    def estimate(self, design_point: List[int], metric: Metric) -> float:
        pass

    @abstractmethod
    def add_example(self, example: Example) -> None:
        pass

    @abstractmethod
    def get_examples(self) -> List[Example]:
        pass


class AbstractDesignPointEvaluator(abc.ABC):
    @abstractmethod
    def evaluate(self, design_point: Tuple[int, ...]) -> DesignValue:
        pass

    @abstractmethod
    def get_max_frequency(self, wns: float) -> float:
        pass

    @abstractmethod
    def set_estimator(self, estimator: AbstractEstimator) -> None:
        pass

    @abstractmethod
    def get_metrics(self):
        pass

    @abstractmethod
    def set_metrics(self, metrics):
        pass


class AbstractFitnessEvaluator(abc.ABC):
    @abstractmethod
    def fitness(self, design_point: Tuple[int, ...]):
        pass


class AbstractSourceParser(abc.ABC):
    @abstractmethod
    def parse(self) -> Tuple[List[Entity], Optional[TopLevel]]:
        pass

    @abstractmethod
    def get_entity(self, entity: str) -> Entity:
        pass

    @abstractmethod
    def set_entity(self, entity: str) -> None:
        pass

    @abstractmethod
    def get_top_entity(self) -> Entity:
        pass

    @abstractmethod
    def get_path(self) -> str:
        pass

    @abstractmethod
    def get_folder(self) -> str:
        pass

    @abstractmethod
    def get_hdl(self) -> RTL:
        pass

    @abstractmethod
    def get_entities(self) -> List[Entity]:
        pass

    @abstractmethod
    def get_parameters(self) -> List[Parameter]:
        pass

    @abstractmethod
    def get_parameter(self, parameter: str) -> Parameter:
        pass

    @abstractmethod
    def get_top_level(self):
        pass

    @abstractmethod
    def get_parameter_value(self, parameter: str) -> Optional[int]:
        pass

    @abstractmethod
    def write_parameter_values(
        self, hdl_handler: FillHandler, values: Dict[str, int],
    ):
        pass

    @abstractmethod
    def get_ports(self) -> List[Port]:
        pass

    @abstractmethod
    def get_port(self, port: str) -> Port:
        pass

    @abstractmethod
    def check_port(self, port: str) -> Optional[Port]:
        pass

    @abstractmethod
    def get_port_direction(self, port: str) -> Optional[str]:
        pass

    @abstractmethod
    def get_port_type(self, port: str) -> Optional[str]:
        pass

    @abstractmethod
    def get_imports(self) -> Tuple[List[str], List[str]]:
        pass




# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['dovado_rtl',
 'dovado_rtl.antlr',
 'dovado_rtl.antlr.generated',
 'dovado_rtl.antlr.generated.SysVerilogHDL',
 'dovado_rtl.antlr.generated.Verilog2001',
 'dovado_rtl.antlr.generated.vhdl',
 'tcl',
 'verilog',
 'vhdl',
 'xdc']

package_data = \
{'': ['*'], 'dovado_rtl.antlr': ['grammars/*', 'grammars/.antlr/*']}

install_requires = \
['BeautifulSoup4>=4.9.1,<5.0.0',
 'antlr4-python3-runtime>=4.8.0,<4.9.0',
 'lxml>=4.5.2,<5.0.0',
 'movado>=0.4.0,<0.5.0',
 'pathvalidate>=2.3.0,<3.0.0',
 'pexpect>=4.8.0,<5.0.0',
 'pymoo==0.4.1',
 'pyyaml>=5.3.1,<6.0.0',
 'typer>=0.3.2,<0.4.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.8,<0.9']}

entry_points = \
{'console_scripts': ['dovado = dovado_rtl.main:main']}

setup_kwargs = {
    'name': 'dovado-rtl',
    'version': '0.7.5',
    'description': 'CLI tool for RTL Design Space Exploration on top of Vivado',
    'long_description': '\n# Table of Contents\n\n1.  [Installation](#orgc020a3e)\n2.  [Usage](#org4f8647b)\n    1.  [Examples](#org12b2aea)\n        1.  [neorv32 (VHDL)](#orgd8de369)\n        2.  [corundum (VERILOG)](#org43556a3)\n        3.  [cv32e40p (SYSTEM-VERILOG)](#org6e54816)\n\nDoVado is a RTL design automation and exploration CLI tool.\n\n\n<a id="orgc020a3e"></a>\n\n# Installation\n\nDoVado needs python 3.6 or higher. Install it through pip, on many Linux systems use pip3 to force python 3 installation.\n\n    pip3 install --user --no-cache dovado-rtl\n\n\n<a id="org4f8647b"></a>\n\n# Usage\n\nDovado has two modes:\n\n-   points: design automation mode in which a file containing parameter values must be given and a file containing all the evaluations is returned for some given metrics,\n-   space: design exploration mode in which parameters and their ranges must be given together with some target metrics and the pareto set of design points with respect to the given metrics is returned.\n\n<table border="2" cellspacing="0" cellpadding="6" rules="groups" frame="hsides">\n<caption class="t-above"><span class="table-number">Table 1:</span> dovado general parameters</caption>\n\n<colgroup>\n<col  class="org-left" />\n\n<col  class="org-left" />\n\n<col  class="org-left" />\n</colgroup>\n<thead>\n<tr>\n<th scope="col" class="org-left">parameter</th>\n<th scope="col" class="org-left">description</th>\n<th scope="col" class="org-left">mandatory</th>\n</tr>\n</thead>\n\n<tbody>\n<tr>\n<td class="org-left">&#x2013;file-path</td>\n<td class="org-left">path to the target file</td>\n<td class="org-left">yes</td>\n</tr>\n\n\n<tr>\n<td class="org-left">&#x2013;board</td>\n<td class="org-left">vivado descriptor of a board</td>\n<td class="org-left">yes</td>\n</tr>\n\n\n<tr>\n<td class="org-left">&#x2013;parameters</td>\n<td class="org-left">parameters to use either for points/space</td>\n<td class="org-left">yes</td>\n</tr>\n\n\n<tr>\n<td class="org-left">&#x2013;clock-port</td>\n<td class="org-left">RTL identifier of the clock port</td>\n<td class="org-left">yes</td>\n</tr>\n\n\n<tr>\n<td class="org-left">&#x2013;implementation</td>\n<td class="org-left">switch to evaluate designs after implementation (default is after synthesis)</td>\n<td class="org-left">no</td>\n</tr>\n\n\n<tr>\n<td class="org-left">&#x2013;incremental</td>\n<td class="org-left">switch to use incremental synthesis/implementation</td>\n<td class="org-left">no</td>\n</tr>\n\n\n<tr>\n<td class="org-left">&#x2013;directives</td>\n<td class="org-left">list of directives to pass to synthesis, place and route (default is RuntimeOptimized for all three)</td>\n<td class="org-left">no</td>\n</tr>\n\n\n<tr>\n<td class="org-left">&#x2013;target-clock</td>\n<td class="org-left">clock (Mhz) to give as a constraint to Vivado (default=1000)</td>\n<td class="org-left">no</td>\n</tr>\n\n\n<tr>\n<td class="org-left">&#x2013;metrics</td>\n<td class="org-left">list of metrics to target using their integer identifier (default mode is interactive, you will be asked after first synthesis/implementation)</td>\n<td class="org-left">no</td>\n</tr>\n</tbody>\n</table>\n\nAfter those parameters specify points/space both these modes take an argument:\n\n-   points argument: specify the path to the csv file containing the design points to be analyzed. The csv file must contain on each line the value for each of the parameters stated through &#x2013;parameters in the same order,\n-   space argument: a list of ranges stated as 1 2 3 4 where this way we would be defining two ranges (1, 2) for the first parameter and (3, 4) for the second parameter\n\nNo further parameters can be passed to points\n\n<table border="2" cellspacing="0" cellpadding="6" rules="groups" frame="hsides">\n<caption class="t-above"><span class="table-number">Table 2:</span> dovado space parameters</caption>\n\n<colgroup>\n<col  class="org-left" />\n\n<col  class="org-left" />\n\n<col  class="org-left" />\n</colgroup>\n<thead>\n<tr>\n<th scope="col" class="org-left">parameter</th>\n<th scope="col" class="org-left">description</th>\n<th scope="col" class="org-left">mandatory</th>\n</tr>\n</thead>\n\n<tbody>\n<tr>\n<td class="org-left">&#x2013;power-of-2</td>\n<td class="org-left">list of &rsquo;y&rsquo; or &rsquo;n&rsquo; where each corresponding parameter, in the same order, is specified as a power of 2 (default is &ldquo;n&rdquo; for all parameters)</td>\n<td class="org-left">no</td>\n</tr>\n\n\n<tr>\n<td class="org-left">&#x2013;param-initial-values</td>\n<td class="org-left">state parameters which are guaranteed to synthesize/implement in order to retrieve the usage metrics after first synthesis</td>\n<td class="org-left">no</td>\n</tr>\n\n\n<tr>\n<td class="org-left">&#x2013;optimization-runtime</td>\n<td class="org-left">set as a termination condition a timeout (which will be taken as a hint and not strictly respected) as hh:mm:ss (default is a tolerance based termination criterion)</td>\n<td class="org-left">no</td>\n</tr>\n\n\n<tr>\n<td class="org-left">&#x2013;record-design-values</td>\n<td class="org-left">record all design values in a csv file</td>\n<td class="org-left">no</td>\n</tr>\n\n\n<tr>\n<td class="org-left">&#x2013;read-design-values</td>\n<td class="org-left">read design values from a csv file</td>\n<td class="org-left">no</td>\n</tr>\n</tbody>\n</table>\n\nDirectory structure is vital for the functioning of the tool:\n\n-   VHDL: if a package is used the corresponding folder must be named exactly as the package; if one wants to analyse a module in a project with multiple packages each file belonging to a given package must reside in a subfolder with the same name as the package it belongs to:\n    -   package-name (top folder must have the name of the top package if it exists or any name if it does not exist)\n        -   file-1 (belonging to package-name)\n        -   file-2 (belonging to package-name)\n        -   subpackage1-name\n            -   file-1 (belonging to subpackage-name)\n            -   file-2 (belonging to subpackage-name)\n            -   &#x2026;\n        -   subpackage2-name\n            -   &#x2026;\n        -   &#x2026;\n-   VERILOG/SYSTEM-VERILOG: include directives are not supported all files must be in the same folder, no subfolders allowed.\n\n\n<a id="org12b2aea"></a>\n\n## Examples\n\n\n<a id="orgd8de369"></a>\n\n### neorv32 (VHDL)\n\n    git clone https://github.com/stnolting/neorv32\n    cd neorv32/rtl\n    mv core neorv32\n\nChanging the name of the core folder, which contains all vhdl files, to the name of the package which is used along the files is mandatory to make dovado get &rsquo;use&rsquo; directives right.\nExploring the parameter space of the top module:\n\n    dovado --file-path <path to "neorv32/rtl/neorv32/neorv32_top.vhd"> --board xc7k70tfbv676-1 --parameters MEM_INT_IMEM_SIZE --parameters MEM_INT_DMEM_SIZE --clock-port clk_i --metrics 0 --metrics 1 --metrics 4 --metrics 9 space 16384 131072 8129 65536 --power-of-2 y --power-of-2 y\n\nAbove we are optimizing two memory parameters (MEM<sub>INT</sub><sub>IMEM</sub><sub>SIZE</sub>, MEM<sub>INT</sub><sub>DMEM</sub><sub>SIZE</sub>) with clk<sub>i</sub> as the clock port with metrics chosen:\n\n-   frequency (0)\n-   LUT occupation (1)\n-   REGISTER occupation (4)\n-   BRAM occupation (9)\n\nRanges are specified after space and we also specify that we want to search only among power of 2&rsquo;s solutions.\n\n\n<a id="org43556a3"></a>\n\n### corundum (VERILOG)\n\n    git clone https://github.com/corundum/corundum\n    cd corundum/\n\nExploring the parameter space of the top module:\n\n    dovado --file-path <path to "corundum/fpga/common/rtl/cpl_queue_manager.v"> --board xc7k70tfbv676-1 --target-clock 100000 --parameters OP_TABLE_SIZE --parameters QUEUE_INDEX_WIDTH --parameters PIPELINE --clock-port clk --metrics 0 --metrics 1 --metrics 4 --metrics 9 space 8 64 4 11 2 32 --record-design-values\n\n\n<a id="org6e54816"></a>\n\n### cv32e40p (SYSTEM-VERILOG)\n\n    git clone https://github.com/openhwgroup/cv32e40p\n    cd rtl\n    mkdir testing\n    cp cv32e40p_fifo.sv testing/\n\nIn this project an include directory is used but dovado does not currently support it thus we create a subfolder, name may be whatever, where to isolate the module we are interested in studying. This workaround is only possible if the module one wants to study works standalone without include directives.\n\n    dovado --file-path ../../test_projects/cv32e40p/rtl/testing/cv32e40p_fifo.sv --board xc7k70tfbv676-1 --target-clock 100000 --parameters DEPTH --parameters DATA_WIDTH --clock-port clk_i --metrics 0 --metrics 1 --metrics 4 --metrics 9 space 2 4294967296 2 64 --power-of-2 y --power-of-2 y\n\n',
    'author': 'Daniele Paletti',
    'author_email': 'danielepaletti98@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/DPaletti/dovado',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
